var EventEmitter, MemorySync, Model, merge, onMixins, pathParser;
var __slice = Array.prototype.slice;
MemorySync = require('./adapters/MemorySync');
pathParser = require('./pathParser');
EventEmitter = require('events').EventEmitter;
merge = require('./util').mergeAll;
Model = module.exports = function(_clientId, AdapterClass) {
  var adapter, init, self, _i, _len, _ref;
  this._clientId = _clientId != null ? _clientId : '';
  if (AdapterClass == null) {
    AdapterClass = MemorySync;
  }
  self = this;
  self._adapter = adapter = new AdapterClass;
  _ref = Model.mixins;
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    init = _ref[_i].init;
    if (init) {
      init.call(self);
    }
  }
  self.silent = Object.create(self, {
    _silent: {
      value: true
    }
  });
};
Model.prototype._setSocket = function(socket) {
  var onConnected, self, setupSocket, _i, _len, _ref, _results;
  self = this;
  self.socket = socket;
  self.canConnect = true;
  socket.on('fatalErr', function() {
    self.canConnect = false;
    self.emit('canConnect', false);
    return socket.disconnect();
  });
  self.connected = false;
  onConnected = function() {
    self.emit('connected', self.connected);
    return self.emit('connectionStatus', self.connected, self.canConnect);
  };
  socket.on('connect', function() {
    self.connected = true;
    return onConnected();
  });
  socket.on('disconnect', function() {
    self.connected = false;
    return setTimeout(onConnected, 200);
  });
  socket.on('connect_failed', onConnected);
  _ref = Model.mixins;
  _results = [];
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    setupSocket = _ref[_i].setupSocket;
    _results.push(setupSocket ? setupSocket.call(this, socket) : void 0);
  }
  return _results;
};
Model.prototype["with"] = function(options) {
  return Object.create(this, {
    _with: {
      value: options
    }
  });
};
merge(Model.prototype, EventEmitter.prototype, {
  _eventListener: function(method, pattern, callback) {
    var re;
    if (pattern.call) {
      return pattern;
    }
    re = pathParser.eventRegExp(pattern);
    return function(_arg, isLocal) {
      var args, path;
      path = _arg[0], args = 2 <= _arg.length ? __slice.call(_arg, 1) : [];
      if (re.test(path)) {
        callback.apply(null, re.exec(path).slice(1).concat(args, isLocal));
        return true;
      }
    };
  },
  _on: EventEmitter.prototype.on,
  on: function(type, pattern, callback) {
    var listener;
    this._on(type, listener = this._eventListener(type, pattern, callback));
    return listener;
  },
  once: function(type, pattern, callback) {
    var g, listener, self;
    listener = this._eventListener(type, pattern, callback);
    self = this;
    this._on(type, g = function() {
      var matches;
      matches = listener.apply(null, arguments);
      if (matches) {
        return self.removeListener(type, g);
      }
    });
    return listener;
  }
});
Model.prototype.addListener = Model.prototype.on;
Model.mixins = [];
Model.accessors = {};
Model.mutators = {};
onMixins = [];
Model.mixin = function(mixin) {
  var cache, category, fn, key, name, obj, objs, onMixin, proto, static, value, _i, _j, _len, _len2, _ref;
  Model.mixins.push(mixin);
  if (proto = mixin.proto) {
    merge(Model.prototype, proto);
  }
  if (static = mixin.static) {
    merge(Model, static);
  }
  _ref = ['accessors', 'mutators'];
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    category = _ref[_i];
    cache = Model[category];
    if (objs = mixin[category]) {
      for (name in objs) {
        obj = objs[name];
        Model.prototype[name] = cache[name] = fn = obj.fn;
        for (key in obj) {
          value = obj[key];
          if (key === 'fn') {
            continue;
          }
          fn[key] = value;
        }
      }
    }
  }
  if (onMixin = mixin.onMixin) {
    onMixins.push(onMixin);
  }
  for (_j = 0, _len2 = onMixins.length; _j < _len2; _j++) {
    onMixin = onMixins[_j];
    onMixin(Model.mutators, Model.accessors);
  }
  return Model;
};
Model.mixin(require('./mixin.subscribe'));
Model.mixin(require('./mixin.refs'));
Model.mixin(require('./mixin.stm'));
Model.mixin(require('./mixin.ot'));