var PubSub, RedisAdapter, handlePaths, hasKeys, pathParser, redis, transaction;
redis = require('redis');
pathParser = require('./pathParser');
transaction = require('./transaction.server');
hasKeys = require('./util').hasKeys;
PubSub = module.exports = function(options) {
  var adapterName, onMessage;
  if (options == null) {
    options = {};
  }
  adapterName = options.adapter || 'Redis';
  onMessage = options.onMessage || function() {};
  this._adapter = new PubSub._adapters[adapterName](onMessage, options);
};
PubSub.prototype = {
  subscribe: function(subscriberId, paths, callback) {
    return this._adapter.subscribe(subscriberId, paths, callback);
  },
  publish: function(path, message) {
    return this._adapter.publish(path, message);
  },
  unsubscribe: function(subscriberId, paths, callback) {
    return this._adapter.unsubscribe(subscriberId, paths, callback);
  },
  hasSubscriptions: function(subscriberId) {
    return this._adapter.hasSubscriptions(subscriberId);
  },
  subscribedToTxn: function(subscriberId, txn) {
    return this._adapter.subscribedToTxn(subscriberId, txn);
  }
};
PubSub._adapters = {};
PubSub._adapters.Redis = RedisAdapter = function(onMessage, options) {
  var db, event, host, makeCallback, namespace, port, redisOptions, subClient, subs, _fn, _i, _len, _ref, _ref2;
  redisOptions = (_ref = options.redis || {}, port = _ref.port, host = _ref.host, db = _ref.db, _ref);
  namespace = (db || 0) + '.';
  this._namespace = function(path) {
    return namespace + path;
  };
  if (!(this._publishClient = options.pubClient)) {
    this._publishClient = redis.createClient(port, host, redisOptions);
    if (db) {
      this._publishClient.select(db);
    }
  }
  if (!(this._subscribeClient = subClient = options.subClient)) {
    this._subscribeClient = redis.createClient(port, host, redisOptions);
  }
  this._subs = subs = {};
  this._subscriberSubs = {};
  if (options.debug) {
    _ref2 = ['subscribe', 'unsubscribe', 'psubscribe', 'punsubscribe'];
    _fn = function(event) {
      return subClient.on(event, function(path, count) {
        return console.log("" + (event.toUpperCase()) + " " + path + " COUNT = " + count);
      });
    };
    for (_i = 0, _len = _ref2.length; _i < _len; _i++) {
      event = _ref2[_i];
      _fn(event);
    }
    subClient.on('message', function(channel, message) {
      return console.log("MESSAGE " + channel + " " + message);
    });
    subClient.on('pmessage', function(pattern, channel, message) {
      return console.log("PMESSAGE " + pattern + " " + channel + " " + message);
    });
    this.__publish = RedisAdapter.prototype.publish;
    this.publish = function(path, message) {
      console.log("PUBLISH " + (this._namespace(path)) + " " + (JSON.stringify(message)));
      return this.__publish(path, message);
    };
  }
  subClient.on('pmessage', function(pattern, path, message) {
    var pathSubs, re, subscriberId, _results;
    if (pathSubs = subs[pattern.substr(0, pattern.length - 1)]) {
      message = JSON.parse(message);
      _results = [];
      for (subscriberId in pathSubs) {
        re = pathSubs[subscriberId];
        _results.push(re.test(path) ? onMessage(subscriberId, message) : void 0);
      }
      return _results;
    }
  });
  makeCallback = function(queue, event) {
    return subClient.on(event, function(path) {
      var callback, pending;
      if (pending = queue[path]) {
        if (callback = pending.shift()) {
          if (!--callback.__count) {
            return callback();
          }
        }
      }
    });
  };
  makeCallback(this._pendingSubscribe = {}, 'psubscribe');
  makeCallback(this._pendingUnsubscribe = {}, 'punsubscribe');
};
RedisAdapter.prototype = {
  subscribe: function(subscriberId, paths, callback) {
    var path, pathSubs, re, ss, subs, subscriberSubs, toAdd, _i, _len;
    if (subscriberId === void 0) {
      return;
    }
    subs = this._subs;
    subscriberSubs = this._subscriberSubs;
    toAdd = [];
    for (_i = 0, _len = paths.length; _i < _len; _i++) {
      path = paths[_i];
      path = this._namespace(path);
      if (!(pathSubs = subs[path])) {
        subs[path] = pathSubs = {};
        toAdd.push(path);
      }
      re = pathParser.regExp(path);
      pathSubs[subscriberId] = re;
      ss = subscriberSubs[subscriberId] || (subscriberSubs[subscriberId] = {});
      ss[path] = re;
    }
    return handlePaths(toAdd, this._pendingSubscribe, this._subscribeClient, 'psubscribe', callback);
  },
  publish: function(path, message) {
    path = this._namespace(path);
    return this._publishClient.publish(path, JSON.stringify(message));
  },
  unsubscribe: function(subscriberId, paths, callback) {
    var path, pathSubs, ss, subs, subscriberSubs, toRemove, _i, _len;
    if (subscriberId === void 0) {
      return;
    }
    if (typeof paths === 'function') {
      callback = paths;
      paths = null;
    }
    subscriberSubs = this._subscriberSubs;
    paths || (paths = subscriberSubs[subscriberId] || []);
    subs = this._subs;
    toRemove = [];
    for (_i = 0, _len = paths.length; _i < _len; _i++) {
      path = paths[_i];
      path = this._namespace(path);
      if (pathSubs = subs[path]) {
        delete pathSubs[subscriberId];
        if (!hasKeys(pathSubs)) {
          toRemove.push(path);
        }
      }
      if (ss = subscriberSubs[subscriberId]) {
        delete ss[path];
      }
    }
    return handlePaths(toRemove, this._pendingUnsubscribe, this._subscribeClient, 'punsubscribe', callback);
  },
  hasSubscriptions: function(subscriberId) {
    return subscriberId in this._subscriberSubs;
  },
  subscribedToTxn: function(subscriberId, txn) {
    var p, path, re, _ref;
    path = this._namespace(transaction.path(txn));
    _ref = this._subscriberSubs[subscriberId];
    for (p in _ref) {
      re = _ref[p];
      if (p === path || re.test(path)) {
        return true;
      }
    }
    return false;
  }
};
handlePaths = function(paths, queue, client, fn, callback) {
  var i, path, _results;
  if (i = paths.length) {
    if (callback) {
      callback.__count = i;
    }
  } else {
    if (callback) {
      callback();
    }
  }
  _results = [];
  while (i--) {
    client[fn](path = paths[i] + '*');
    _results.push(callback ? (queue[path] || (queue[path] = [])).push(callback) : void 0);
  }
  return _results;
};