var COMMIT, LOCK, LOCK_CLOCK_MASK, LOCK_TIMEOUT, LOCK_TIMEOUT_MASK, MAX_RETRIES, RETRY_DELAY, Stm, UNLOCK, transaction;
var __slice = Array.prototype.slice;
transaction = require('./transaction.server');
MAX_RETRIES = 10;
RETRY_DELAY = 5;
Stm = module.exports = function(redisClient) {
  var getLocks, lock, lockQueue;
  lockQueue = {};
  lock = function(numKeys, locks, sinceVer, path, retries, delay, callback) {
    return redisClient.eval.apply(redisClient, [LOCK, numKeys].concat(__slice.call(locks), [sinceVer], [function(err, values) {
      var queue;
      if (err) {
        return callback(err);
      }
      if (values[0]) {
        return callback(null, values[0], values[1]);
      }
      if (retries) {
        queue = lockQueue[path] || (lockQueue[path] = []);
        queue.push([numKeys, locks, sinceVer, path, retries - 1, delay * 2, callback]);
        return setTimeout(function() {
          var args;
          if (args = lockQueue[path].shift()) {
            return lock.apply(null, args);
          }
        }, delay);
      }
      return callback('lockMaxRetries');
    }]));
  };
  this._getLocks = getLocks = function(path) {
    var lockPath, segment;
    lockPath = '';
    return ((function() {
      var _i, _len, _ref, _results;
      _ref = path.split('.');
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        segment = _ref[_i];
        _results.push(lockPath += '.' + segment);
      }
      return _results;
    })()).reverse();
  };
  this.commit = function(txn, callback) {
    var base, locks, locksLen, paths, sinceVer;
    base = transaction.base(txn);
    sinceVer = base == null ? '' : base + 1;
    if (transaction.isCompound(txn)) {
      paths = transaction.ops(txn).map(function(op) {
        return transaction.op.path(op);
      });
    } else {
      paths = [transaction.path(txn)];
    }
    locks = paths.reduce(function(locks, path) {
      getLocks(path).forEach(function(lock) {
        if (-1 === locks.indexOf(lock)) {
          return locks.push(lock);
        }
      });
      return locks;
    }, []);
    locksLen = locks.length;
    return lock(locksLen, locks, sinceVer, paths, MAX_RETRIES, RETRY_DELAY, function(err, lockVal, txns) {
      var conflict, path;
      path = paths[0];
      if (err) {
        return callback(err);
      }
      if (txns && (conflict = transaction.journalConflict(txn, txns))) {
        return redisClient.eval.apply(redisClient, [UNLOCK, locksLen].concat(__slice.call(locks), [lockVal], [function(err) {
          if (err) {
            return callback(err);
          }
          return callback(conflict);
        }]));
      }
      return redisClient.eval.apply(redisClient, [COMMIT, locksLen].concat(__slice.call(locks), [lockVal], [JSON.stringify(txn)], [function(err, ver) {
        var args, queue;
        if (err) {
          return callback(err);
        }
        if (ver === 0) {
          return callback('lockReleased');
        }
        callback(null, ver);
        if ((queue = lockQueue[path]) && (args = queue.shift())) {
          return lock.apply(null, args);
        }
      }]));
    });
  };
};
Stm._LOCK_TIMEOUT = LOCK_TIMEOUT = 3;
Stm._LOCK_TIMEOUT_MASK = LOCK_TIMEOUT_MASK = 0x100000000;
Stm._LOCK_CLOCK_MASK = LOCK_CLOCK_MASK = 0x100000;
Stm._LOCK = LOCK = "local now = os.time()\nlocal path = KEYS[1]\nfor i, lock in pairs(redis.call('smembers', path)) do\n  if lock % " + LOCK_TIMEOUT_MASK + " < now then\n    redis.call('srem', path, lock)\n  else\n    return 0\n  end\nend\nfor i, path in pairs(KEYS) do\n  path = 'l' .. path\n  local val = redis.call('get', path)\n  if val then\n    if val % " + LOCK_TIMEOUT_MASK + " < now then\n      redis.call('del', path)\n    else\n      return 0\n    end\n  end\nend\nlocal lock = '0x' ..\n  string.format('%x', redis.call('incr', 'lockClock') % " + LOCK_CLOCK_MASK + ") ..\n  string.format('%x', now + " + LOCK_TIMEOUT + ")\nredis.call('set', 'l' .. path, lock)\nfor i, path in pairs(KEYS) do\n  redis.call('sadd', path, lock)\nend\nlocal txns\nif ARGV[1] ~= '' then txns = redis.call('zrangebyscore', 'txns', ARGV[1], '+inf') end\nreturn {lock, txns}";
Stm._UNLOCK = UNLOCK = "local val = ARGV[1]\nlocal path = 'l' .. KEYS[1]\nif redis.call('get', path) == val then redis.call('del', path) end\nfor i, path in pairs(KEYS) do\n  redis.call('srem', path, val)\nend";
Stm._COMMIT = COMMIT = "local val = ARGV[1]\nlocal path = 'l' .. KEYS[1]\nlocal fail = false\nif redis.call('get', path) == val then redis.call('del', path) else fail = true end\nfor i, path in pairs(KEYS) do\n  if redis.call('srem', path, val) == 0 then return 0 end\nend\nif fail then return 0 end\nlocal ver = redis.call('incr', 'ver')\nredis.call('zadd', 'txns', ver, ARGV[2])\nreturn ver";