var MemorySync, lookup, lookupAddPath, lookupSetVersion, lookupWithVersion, specClone, _ref;
var __slice = Array.prototype.slice;
_ref = require('./lookup'), lookup = _ref.lookup, lookupWithVersion = _ref.lookupWithVersion, lookupAddPath = _ref.lookupAddPath, lookupSetVersion = _ref.lookupSetVersion;
specClone = require('../specHelper').clone;
MemorySync = module.exports = function() {
  this._data = {
    world: {}
  };
  this._vers = {
    ver: 0
  };
};
MemorySync.prototype = {
  version: function(path, data) {
    if (path) {
      return lookupWithVersion(path, data || this._data, this._vers)[1].ver;
    } else {
      return this._vers.ver;
    }
  },
  get: function(path, data) {
    data || (data = this._data);
    if (path) {
      return lookup(path, data);
    } else {
      return data.world;
    }
  },
  getWithVersion: function(path, data) {
    var currVer, obj, _ref2;
    data || (data = this._data);
    if (path) {
      _ref2 = lookupWithVersion(path, data, this._vers), obj = _ref2[0], currVer = _ref2[1];
      return [obj, currVer.ver];
    } else {
      return [data.world, this._vers.ver];
    }
  },
  getRef: function(path, data) {
    return lookup(path, data || this._data, true);
  },
  getAddPath: function(path, data, ver, pathType) {
    return lookupAddPath(path, data || this._data, ver, pathType);
  },
  set: function(path, value, ver, data) {
    var parent, prop, _ref2;
    _ref2 = lookupSetVersion(path, data || this._data, this._vers, ver, 'object'), parent = _ref2[1], prop = _ref2[2];
    return parent[prop] = value;
  },
  del: function(path, ver, data) {
    var grandparent, index, obj, parent, parentClone, parentPath, parentProp, prop, _ref2, _ref3;
    data || (data = this._data);
    _ref2 = lookupSetVersion(path, data, this._vers, ver), obj = _ref2[0], parent = _ref2[1], prop = _ref2[2];
    if (ver != null) {
      delete parent[prop];
      return obj;
    }
    if (!parent) {
      return obj;
    }
    if (~(index = path.lastIndexOf('.'))) {
      parentPath = path.substr(0, index);
      _ref3 = lookupSetVersion(parentPath, data, this._vers, ver), parent = _ref3[0], grandparent = _ref3[1], parentProp = _ref3[2];
    } else {
      parent = data.world;
      grandparent = data;
      parentProp = 'world';
    }
    parentClone = specClone(parent);
    delete parentClone[prop];
    grandparent[parentProp] = parentClone;
    return obj;
  },
  push: function() {
    var args, arr, data, path, ver, _i;
    path = arguments[0], args = 4 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 2) : (_i = 1, []), ver = arguments[_i++], data = arguments[_i++];
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    return arr.push.apply(arr, args);
  },
  unshift: function() {
    var args, arr, data, path, ver, _i;
    path = arguments[0], args = 4 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 2) : (_i = 1, []), ver = arguments[_i++], data = arguments[_i++];
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    return arr.unshift.apply(arr, args);
  },
  splice: function() {
    var args, arr, data, path, ver, _i;
    path = arguments[0], args = 4 <= arguments.length ? __slice.call(arguments, 1, _i = arguments.length - 2) : (_i = 1, []), ver = arguments[_i++], data = arguments[_i++];
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    return arr.splice.apply(arr, args);
  },
  pop: function(path, ver, data) {
    var arr;
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    return arr.pop();
  },
  shift: function(path, ver, data) {
    var arr;
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    return arr.shift();
  },
  insertAfter: function(path, afterIndex, value, ver, data) {
    var arr, outOfBounds;
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    outOfBounds = !((-1 <= afterIndex && afterIndex <= arr.length - 1));
    if (outOfBounds) {
      throw new Error('Out of Bounds');
    }
    arr.splice(afterIndex + 1, 0, value);
    return arr.length;
  },
  insertBefore: function(path, beforeIndex, value, ver, data) {
    var arr, outOfBounds;
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    outOfBounds = !((0 <= beforeIndex && beforeIndex <= arr.length));
    if (outOfBounds) {
      throw new Error('Out of Bounds');
    }
    arr.splice(beforeIndex, 0, value);
    return arr.length;
  },
  remove: function(path, startIndex, howMany, ver, data) {
    var arr, outOfBounds;
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    outOfBounds = !((0 <= startIndex && startIndex <= (arr.length && arr.length - 1 || 0)));
    if (outOfBounds) {
      throw new Error('Out of Bounds');
    }
    return arr.splice(startIndex, howMany);
  },
  move: function(path, from, to, ver, data) {
    var arr, len, outOfBounds, value;
    arr = lookupSetVersion(path, data || this._data, this._vers, ver, 'array')[0];
    if (!Array.isArray(arr)) {
      throw new Error('Not an Array');
    }
    len = arr.length;
    if (from < 0) {
      from += len;
    }
    if (to < 0) {
      to += len;
    }
    outOfBounds = !(((0 <= from && from < len)) && ((0 <= to && to < len)));
    if (outOfBounds) {
      throw new Error('Out of Bounds');
    }
    value = arr.splice(from, 1)[0];
    arr.splice(to, 0, value);
    return value;
  }
};