var Async, AtomicModel, RESEND_INTERVAL, SEND_TIMEOUT, Serializer, pathParser, specCreate, stm, transaction;
var __slice = Array.prototype.slice;
transaction = require('../transaction');
pathParser = require('../pathParser');
Serializer = require('../Serializer');
specCreate = require('../specHelper').create;
AtomicModel = require('./AtomicModel');
Async = require('./Async');
stm = module.exports = {
  static: {
    _SEND_TIMEOUT: SEND_TIMEOUT = 10000,
    _RESEND_INTERVAL: RESEND_INTERVAL = 2000
  },
  init: function() {
    var adapter, self, txnApplier, txnQueue, txns;
    this._specCache = {
      invalidate: function() {
        delete this.data;
        return delete this.lastTxnId;
      }
    };
    this._startId = '';
    this._atomicModels = {};
    this._txnCount = 0;
    this._txns = txns = {};
    this._txnQueue = txnQueue = [];
    adapter = this._adapter;
    self = this;
    txnApplier = new Serializer({
      withEach: function(txn) {
        if (transaction.base(txn) > adapter.version()) {
          return self._applyTxn(txn);
        }
      },
      onTimeout: function() {
        return self._reqNewTxns();
      }
    });
    this._onTxn = function(txn, num) {
      var queuedTxn;
      if (queuedTxn = txns[transaction.id(txn)]) {
        txn.callback = queuedTxn.callback;
        txn.emitted = queuedTxn.emitted;
      }
      return txnApplier.add(txn, num);
    };
    this._onTxnNum = function(num) {
      txnApplier.setIndex((+num || 0) + 1);
      return txnApplier.clearPending();
    };
    this._removeTxn = function(txnId) {
      var i;
      delete txns[txnId];
      if (~(i = txnQueue.indexOf(txnId))) {
        txnQueue.splice(i, 1);
      }
      return self._specCache.invalidate();
    };
    this.force = Object.create(this, {
      _force: {
        value: true
      }
    });
    return this.async = new Async(this);
  },
  setupSocket: function(socket) {
    var commit, resend, resendInterval, self, _adapter, _onTxn, _ref, _removeTxn, _txnQueue, _txns;
    _ref = self = this, _adapter = _ref._adapter, _txns = _ref._txns, _txnQueue = _ref._txnQueue, _onTxn = _ref._onTxn, _removeTxn = _ref._removeTxn;
    this._commit = commit = function(txn) {
      if (!socket.socket.connected) {
        return;
      }
      txn.timeout = +new Date + SEND_TIMEOUT;
      return socket.emit('txn', txn, self._startId);
    };
    socket.on('txn', _onTxn);
    socket.on('txnNum', this._onTxnNum);
    socket.on('txnOk', function(txnId, base, num) {
      var txn;
      if (!(txn = _txns[txnId])) {
        return;
      }
      transaction.base(txn, base);
      return _onTxn(txn, num);
    });
    socket.on('txnErr', function(err, txnId) {
      var callback, callbackArgs, txn;
      txn = _txns[txnId];
      if (txn && (callback = txn.callback)) {
        if (transaction.isCompound(txn)) {
          callbackArgs = transaction.ops(txn);
        } else {
          callbackArgs = transaction.args(txn).slice(0);
        }
        callbackArgs.unshift(err);
        callback.apply(null, callbackArgs);
      }
      return _removeTxn(txnId);
    });
    this._reqNewTxns = function() {
      return socket.emit('txnsSince', _adapter.version() + 1, self._startId);
    };
    resendInterval = null;
    resend = function() {
      var id, now, txn, _i, _len, _results;
      now = +new Date;
      _results = [];
      for (_i = 0, _len = _txnQueue.length; _i < _len; _i++) {
        id = _txnQueue[_i];
        txn = _txns[id];
        if (txn.timeout > now) {
          return;
        }
        _results.push(commit(txn));
      }
      return _results;
    };
    socket.on('connect', function() {
      var id, _i, _len;
      for (_i = 0, _len = _txnQueue.length; _i < _len; _i++) {
        id = _txnQueue[_i];
        commit(_txns[id]);
      }
      if (!resendInterval) {
        return resendInterval = setInterval(resend, RESEND_INTERVAL);
      }
    });
    return socket.on('disconnect', function() {
      if (resendInterval) {
        clearInterval(resendInterval);
      }
      return resendInterval = null;
    });
  },
  proto: {
    _commit: function() {},
    _reqNewTxns: function() {},
    _nextTxnId: function() {
      return this._clientId + '.' + this._txnCount++;
    },
    _queueTxn: function(txn, callback) {
      var id;
      txn.callback = callback;
      id = transaction.id(txn);
      this._txns[id] = txn;
      return this._txnQueue.push(id);
    },
    _getVer: function() {
      if (this._force) {
        return null;
      } else {
        return this._adapter.version();
      }
    },
    _addOpAsTxn: function(method, args, callback) {
      var base, id, meta, path, txn;
      this.emit('beforeTxn', method, args);
      base = this._getVer();
      id = this._nextTxnId();
      meta = args.meta;
      txn = transaction.create({
        base: base,
        id: id,
        method: method,
        args: args,
        meta: meta
      });
      this._queueTxn(txn, callback);
      if ((path = args[0]) == null) {
        return;
      }
      if (pathParser.isPrivate(path)) {
        this._specCache.invalidate();
        return this._applyTxn(txn);
      }
      if (!this._silent) {
        args = args.slice();
        this.emit(method + 'Post', args, null, null, meta);
        if ('_with' in this) {
          args = args.concat(this._with);
        }
        this.emit(method, args, true, meta);
      }
      txn.emitted = true;
      this._commit(txn);
      return this._specModel().$out;
    },
    _applyTxn: function(txn) {
      var callback, data, doEmit, isCompound, isLocal, op, ops, out, ver, _i, _len;
      data = this._adapter._data;
      doEmit = !(txn.emitted || this._silent);
      isLocal = 'callback' in txn;
      ver = transaction.base(txn);
      if (isCompound = transaction.isCompound(txn)) {
        ops = transaction.ops(txn);
        for (_i = 0, _len = ops.length; _i < _len; _i++) {
          op = ops[_i];
          this._applyMutation(transaction.op, op, ver, data, doEmit, isLocal);
        }
      } else {
        out = this._applyMutation(transaction, txn, ver, data, doEmit, isLocal);
      }
      this._removeTxn(transaction.id(txn));
      if (callback = txn.callback) {
        if (isCompound) {
          callback.apply(null, [null].concat(__slice.call(transaction.ops(txn))));
        } else {
          callback.apply(null, [null].concat(__slice.call(transaction.args(txn)), [out]));
        }
      }
      return out;
    },
    _applyMutation: function(extractor, mutation, ver, data, doEmit, isLocal) {
      var args, meta, method, obj, _ref;
      method = extractor.method(mutation);
      if (method === 'get') {
        return;
      }
      args = extractor.args(mutation);
      meta = extractor.meta(mutation);
      this.emit(method + 'Pre', args, ver, data, meta);
      obj = (_ref = this._adapter)[method].apply(_ref, __slice.call(args).concat([ver], [data]));
      this.emit(method + 'Post', args, ver, data, meta);
      if (doEmit) {
        this.emit(method, args, isLocal, meta);
      }
      return obj;
    },
    _specModel: function() {
      var cache, data, i, lastTxnId, len, op, ops, out, replayFrom, txn, _i, _len;
      if (!(len = this._txnQueue.length)) {
        return this._adapter._data;
      }
      cache = this._specCache;
      if (lastTxnId = cache.lastTxnId) {
        if (cache.lastTxnId === this._txnQueue[len - 1]) {
          return cache.data;
        }
        data = cache.data;
        replayFrom = 1 + this._txnQueue.indexOf(cache.lastTxnId);
      } else {
        replayFrom = 0;
      }
      if (!data) {
        data = cache.data = specCreate(this._adapter._data);
      }
      i = replayFrom;
      while (i < len) {
        txn = this._txns[this._txnQueue[i++]];
        if (transaction.isCompound(txn)) {
          ops = transaction.ops(txn);
          for (_i = 0, _len = ops.length; _i < _len; _i++) {
            op = ops[_i];
            this._applyMutation(transaction.op, op, null, data);
          }
        } else {
          out = this._applyMutation(transaction, txn, null, data);
        }
      }
      cache.data = data;
      cache.lastTxnId = transaction.id(txn);
      data.$out = out;
      return data;
    },
    snapshot: function() {
      var model;
      model = new AtomicModel(this._nextTxnId(), this);
      model._adapter = adapter.snapshot();
      return model;
    },
    atomic: function(block, callback) {
      var abort, commit, model, retry, self;
      model = new AtomicModel(this._nextTxnId(), this);
      this._atomicModels[model.id] = model;
      self = this;
      commit = function(_callback) {
        return model.commit(function(err) {
          if (!err) {
            delete self._atomicModels[model.id];
          }
          if (_callback || (_callback = callback)) {
            return _callback.apply(null, arguments);
          }
        });
      };
      abort = function() {};
      retry = function() {};
      if (block.length === 1) {
        block(model);
        return commit(callback);
      } else if (block.length === 2) {
        return block(model, commit);
      } else if (block.length === 3) {
        return block(model, commit, abort);
      } else if (block.length === 4) {
        return block(model, commit, abort, retry);
      }
    }
  },
  accessors: {
    get: {
      type: 'basic',
      fn: function(path) {
        return this._adapter.get(path, this._specModel());
      }
    }
  },
  mutators: {
    set: {
      type: 'basic',
      fn: function(path, val, callback) {
        return this._addOpAsTxn('set', [path, val], callback);
      }
    },
    del: {
      type: 'basic',
      fn: function(path, callback) {
        return this._addOpAsTxn('del', [path], callback);
      }
    },
    setNull: {
      type: 'compound',
      fn: function(path, value, callback) {
        var obj;
        obj = this.get(path);
        if (obj != null) {
          return obj;
        }
        return this.set(path, value, callback);
      }
    },
    incr: {
      type: 'compound',
      fn: function(path, byNum, callback) {
        if (typeof byNum === 'function') {
          callback = byNum;
          byNum = 1;
        } else if (typeof byNum !== 'number') {
          byNum = 1;
        }
        return this.set(path, (this.get(path) || 0) + byNum, callback);
      }
    },
    push: {
      type: 'array',
      insertArgs: 1,
      fn: function() {
        var args, callback, _i;
        args = 2 <= arguments.length ? __slice.call(arguments, 0, _i = arguments.length - 1) : (_i = 0, []), callback = arguments[_i++];
        if (typeof callback !== 'function') {
          args.push(callback);
          callback = null;
        }
        return this._addOpAsTxn('push', args, callback);
      }
    },
    unshift: {
      type: 'array',
      insertArgs: 1,
      fn: function() {
        var args, callback, _i;
        args = 2 <= arguments.length ? __slice.call(arguments, 0, _i = arguments.length - 1) : (_i = 0, []), callback = arguments[_i++];
        if (typeof callback !== 'function') {
          args.push(callback);
          callback = null;
        }
        return this._addOpAsTxn('unshift', args, callback);
      }
    },
    splice: {
      type: 'array',
      indexArgs: [1],
      insertArgs: 3,
      fn: function() {
        var args, callback, _i;
        args = 2 <= arguments.length ? __slice.call(arguments, 0, _i = arguments.length - 1) : (_i = 0, []), callback = arguments[_i++];
        if (typeof callback !== 'function') {
          args.push(callback);
          callback = null;
        }
        return this._addOpAsTxn('splice', args, callback);
      }
    },
    pop: {
      type: 'array',
      fn: function(path, callback) {
        return this._addOpAsTxn('pop', [path], callback);
      }
    },
    shift: {
      type: 'array',
      fn: function(path, callback) {
        return this._addOpAsTxn('shift', [path], callback);
      }
    },
    insertBefore: {
      type: 'array',
      indexArgs: [1],
      insertArgs: 2,
      fn: function(path, beforeIndex, value, callback) {
        return this._addOpAsTxn('insertBefore', [path, beforeIndex, value], callback);
      }
    },
    insertAfter: {
      type: 'array',
      indexArgs: [1],
      insertArgs: 2,
      fn: function(path, afterIndex, value, callback) {
        return this._addOpAsTxn('insertAfter', [path, afterIndex, value], callback);
      }
    },
    remove: {
      type: 'array',
      indexArgs: [1],
      fn: function(path, start, howMany, callback) {
        if (typeof howMany === 'function') {
          callback = howMany;
          howMany = 1;
        } else if (typeof howMany !== 'number') {
          howMany = 1;
        }
        return this._addOpAsTxn('remove', [path, start, howMany], callback);
      }
    },
    move: {
      type: 'array',
      indexArgs: [1, 2],
      fn: function(path, from, to, callback) {
        return this._addOpAsTxn('move', [path, from, to], callback);
      }
    }
  }
};